/********************************************************************* 
-- S+ sparse GE solver (version 1.1) --

Copyright (C) 1999 by University of California at Santa Barbara. 
Copyright (C) 2005 by University of Rochester. 
All rights reserved.

S+ team contact: Kai Shen <kshen@cs.rochester.edu>, 
                 Tao Yang <tyang@cs.ucsb.edu>  

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose, without fee, and without written agreement
is hereby granted, provided that the above copyright notice and the
following paragraph appear in all copies of this software.

This software is provided ``as is'', without any warranty, including 
any implied warranty for merchantability or fitness for a particular
purpose. Under no circumstances shall the University of California at 
Santa Barbara, the University of Rochester, or their agents be liable 
for any use of, misuse of, or inability to use this software, including 
incidental and consequential damages.
************************************************************************/

/*
 * LUpFac.c - LU factorization with partial pivoting
 *
 */


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <strings.h>

#include "struct.h"
#include "functions.h"
#include "globals.h"

ELEMTYPE  *SP_Rbuffer, *SP_Cbuffer;		/* element buffers for row and col broadcastings */
INDEXTYPE *SP_Cindexbuf, *SP_Rindexbuf;		/* index buffers for row and col broadcastings*/
PATTYPE   *SP_Cpatternbuf, *SP_Rpatternbuf;	/* pattern buffers for row and col broadcastings*/

double SP_PivotThresh = 1.0;                    /* threshold for threshold pivoting, <=1.0 */

/* SP_PIVOTING_COLBYCOL - column-by-column  pivoting; 
   SP_PIVOTING_LD - batch pivot with large diagnoals; 
   SP_PIVOTING_SBP - batch pivot with speculative factorization */
int SP_BatchPivot = SP_PIVOTING_COLBYCOL;
int SP_BatchPivotBailout = 0;                   /* the number of times batch pivoting has to bail out */

int SP_cbufsize=0, SP_rbufsize=0; 
int SP_cindbufsize=0, SP_rindbufsize=0;
int SP_ibufsize=0;

rowID *SP_PivotVector;

/* fortran char descriptors, required by calling blas subroutines */
_fcd SP_ftcL, SP_ftcU, SP_ftcN;

static ELEMTYPE *_Pbuffer;        /* buffer for pivoting candidates */
static ELEMTYPE *_Bkkbuffer;      /* buffer for bkk and L inverse */
static int *_SwapCounts;	  /* number of swapped rows with each processor */

static void _InitLUP(SP_SNmatrix_struct *SNma) {
  int pbufsize, temp;

  SP_ftcN = _cptofcd("N", strlen("N"));
  SP_ftcU = _cptofcd("U", strlen("U"));
  SP_ftcL = _cptofcd("L", strlen("L"));

  MPI_Allreduce(&SP_cbufsize, &temp, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
  SP_cbufsize = temp;
  MPI_Allreduce(&SP_rbufsize, &temp, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
  SP_rbufsize = temp;
  MPI_Allreduce(&SP_ibufsize, &temp, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
  SP_ibufsize = temp;

  SP_cindbufsize = SNma->snOrder/SP_rp+1;
  SP_rindbufsize = SNma->snOrder/SP_cp+1;

  /* SP_Rbuffer is used to hold the U blocks to do updating */
  SP_Rbuffer = (ELEMTYPE *) SP_Malloc(SP_rbufsize*sizeof(ELEMTYPE));
  SP_Rindexbuf = (INDEXTYPE *) SP_Malloc(SP_rindbufsize*sizeof(INDEXTYPE));
  SP_Rpatternbuf = (PATTYPE *) SP_Malloc(SP_rindbufsize*sizeof(PATTYPE));

  /* SP_Cbuffer is used to hold the L blocks to do updating */
  SP_Cbuffer = (ELEMTYPE *) SP_Malloc(SP_cbufsize*sizeof(ELEMTYPE));
  SP_Cindexbuf = (INDEXTYPE *) SP_Malloc(SP_cindbufsize*sizeof(INDEXTYPE));
  SP_Cpatternbuf = (PATTYPE *) SP_Malloc(SP_cindbufsize*sizeof(PATTYPE));

  /* SP_PivotVector is used to hold the Pivot Vector */
  SP_PivotVector = (rowID*) SP_Malloc((SNma->eleOrder+1) * sizeof(rowID));

  /* _Pbuffer is for holding the pivoting candidates. */
  /* block_no, row_no, and a row of elements */
  if (SP_BatchPivot != SP_PIVOTING_COLBYCOL) 
    pbufsize = SP_rp*SP_BlkSize*(SP_BlkSize+2);
  else 
    pbufsize = SP_rp*(SP_BlkSize+2);
  _Pbuffer = (ELEMTYPE *) SP_Malloc(pbufsize*sizeof(ELEMTYPE));
  bzero(_Pbuffer, pbufsize*sizeof(ELEMTYPE));

  _Bkkbuffer = (ELEMTYPE*) SP_Malloc(SP_BlkSize*SP_BlkSize*sizeof(ELEMTYPE));

  _SwapCounts = (int*) SP_Malloc(SP_rp*sizeof(int));

  SP_InitSblas();	/* allocate temp space for Sblas operations */
  SP_InitComm();	/* allocate temp space for Communication operations */
  SP_InitInterchange();	/* allocate temp space for Interchange operations */
}

static void _FinalLUP() {
  SP_FinalSblas();	/* free temp space for Sblas operations */
  SP_FinalComm();	/* free temp space for Communication operations */
  SP_FinalInterchange();/* free temp space for Interchange operations */

  SP_FreeSize(SP_Rbuffer, SP_rbufsize*sizeof(ELEMTYPE));
  SP_FreeSize(SP_Rindexbuf, SP_rindbufsize*sizeof(INDEXTYPE));
  SP_FreeSize(SP_Rpatternbuf, SP_rindbufsize*sizeof(PATTYPE));
  SP_FreeSize(SP_Cbuffer, SP_cbufsize*sizeof(ELEMTYPE));
  SP_FreeSize(SP_Cindexbuf, SP_cindbufsize*sizeof(INDEXTYPE));
  SP_FreeSize(SP_Cpatternbuf, SP_cindbufsize*sizeof(PATTYPE));

  SP_Free(_Pbuffer);
  SP_Free(_Bkkbuffer);
  SP_Free(_SwapCounts);
}

static void _Factor_SerialPivot (SP_SNmatrix_struct *SNma, INDEXTYPE k) {
  MPI_Status SP_status;
  int ppe, perow, pstart;
  int i, l, l2, m, n, lda, bcol;
  INDEXTYPE stc, endc;
  ELEMTYPE *lbuf, *gbuf;
  int prowreal = -1, prowsize = -1;
  ELEMTYPE alpha;
  ELEMTYPE px, py;
  ELEMTYPE *bkk = NULL, *pa, *bik;

  ppe=PEOWNER(k,k);
  perow=ROWNER(k);
  pstart=COWNER(k)*SP_rp;               /* the pe no of the start of pecol */

  lda = DSIZE (SNma, k);
  stc = SNma->LstartCol[CINDEX(k)];
  endc = SNma->LstartCol[CINDEX(k)+1];
  if (SP_me_no == ppe) bkk = SNma->LnodeCol[stc++];

  /* lbuf is for holding the local pivoting candidates. 
     block_no of local maxima, row_no inside the block of local maxima, 
     and elements for the row containing the local maxima */
  lbuf = _Pbuffer+SP_me_rno*(SP_BlkSize+2);

  for (bcol = 0; bcol < lda; ++bcol) {   /* look at each column in the block */
    px = 0.0;                            /* local maxima */
    pa = NULL;
    if (SP_me_no == ppe) {               /* the pe that has the blk_kk */
      for (i = bcol; i < lda; ++i) {     /* find the max in blk_kk */
        py = fabs (bkk[bcol * lda + i]);
        if (py > px) {
          px = py;
	  *lbuf = k;
	  *(lbuf+1) = i;
          pa = bkk;
        }
      }
      prowreal = *(lbuf+1);
      prowsize = lda;
    }

    for (i = stc; i < endc; ++i) {      /* pivot searching */
      if ((bik = SNma->LnodeCol[i])==NULL)
	continue;

      for (l2 = l = 0; l < DSIZE (SNma, SNma->LindexCol[i]); ++l) {
        if (testcol(SNma->LsubRow[i], l)) {
          py = fabs (bik[bcol * SNma->LsubRowNumInBlk[i] + l2]);
          if (py > px) {
            px = py;
            *lbuf = SNma->LindexCol[i];
            *(lbuf+1) = l;
            pa = bik;
            prowreal = l2;
            prowsize = SNma->LsubRowNumInBlk[i];
          }
          l2 ++;
        }
      }
    }

    /* copy the row containing the local maxima to the buffer */
    if (pa!=NULL && pa[bcol*prowsize+prowreal]!=0.0)  /* find some local maxima */
      for (i=0; i<lda; ++i) (lbuf+2)[i]=pa[i*prowsize+prowreal];
    else {
      *lbuf = SNma->snOrder+1;
      (lbuf+2)[bcol]=0.0; /* make sure this row will not be selected */
    }

    /* gather all local maxima at ppe (owner of bkk) */
    if(SP_me_no == ppe) {
      for (i=0; i<SP_rp; ++i) 
        if (i != SP_me_rno) {
	  MPI_Recv(_Pbuffer+i*(SP_BlkSize+2), lda+2, ELEMTYPE_MPI, pstart+i, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
        }
    } 
    else {  /* not ppe */
      MPI_Send(lbuf, lda+2, ELEMTYPE_MPI, ppe, SP_MPI_TAG, MPI_COMM_WORLD );
    }

    /* determine pivot rows at ppe */
    if(SP_me_no == ppe) {
      int maxproc = -1;
      int minblk = -1;
      px = 0.0;
      for (i=0; i<SP_rp; ++i) { /* scan for all local maximas */
	ELEMTYPE *_buf = _Pbuffer+i*(SP_BlkSize+2);
	py = fabs((_buf+2)[bcol]);     /* the local maxima from pe i */
	if (i == SP_me_rno) py /= SP_PivotThresh;
	                               /* we'd like to find a pivot row in the diagonal
	                                  processor to avoid row swapping, as long as 
	                                  the maximum candidate pivot in the diagonal 
	                                  processor is not too small compared with the 
	                                  global maximum pivot.
	                                  we use threshold pivoting here */
	if (py > px || (py==px && *_buf<minblk))
	  { px=py; maxproc=i; minblk=*_buf;}
      }
      if (px<SP_Epsilon)
	SP_err(SP_ERR_LOC, "Could not find pivot for k=%ld: matrix is singular or EPSILON is too large", k);

      /* get ready to the pivot row broadcast */
      for (i=0; i<SP_rp; ++i)
        if (i!=maxproc) 
	  memcpy(_Pbuffer+i*(SP_BlkSize+2), _Pbuffer+maxproc*(SP_BlkSize+2), (lda+2)*sizeof(ELEMTYPE));

      if (maxproc == SP_me_rno) { /* global maxima is in ppe - local swap */
	ASSERT(pa!=NULL);
        for (i=0; i<lda; ++i) { /* interchange rows */
          py = bkk[lda*i+bcol];
          bkk[lda*i+bcol] = pa[prowsize*i+prowreal];
	  pa[prowsize*i+prowreal] = py;
        }
      }
      else { /* global maxima is not in ppe - swap row with the buf row */
        ELEMTYPE *_buf = _Pbuffer+maxproc*(SP_BlkSize+2);
        for (i=0; i<lda; ++i) {
           py = bkk[lda*i+bcol];
           bkk[lda*i+bcol] = (_buf+2)[i];
           (_buf+2)[i] = py;
        }
      }
    }

    /* broadcast the pivot row or the swapped row from ppe (owner of bkk) */
    if (SP_me_no == ppe) {
      for (i=0; i<SP_rp; ++i)
	if (i != SP_me_rno) 
	  MPI_Send(_Pbuffer+i*(SP_BlkSize+2), lda+2, ELEMTYPE_MPI, pstart+i, SP_MPI_TAG, MPI_COMM_WORLD );
    }
    else {  /* not ppe */
      ELEMTYPE *_buf = _Pbuffer+perow*(SP_BlkSize+2);
      MPI_Recv(_buf, lda+2, ELEMTYPE_MPI, ppe, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);

      if(pa !=NULL && *_buf == *lbuf){ /* this pe has the maximum element */
        for (i=0; i<lda; ++i) { /* interchange row with the buf row */
	  py = pa[prowsize*i+prowreal]; 
          pa[prowsize*i+prowreal] = (_buf+2)[i];
	  (_buf+2)[i] = py; 
        }  
      }
    }

    /* gbuf contains the pivot row */
    gbuf = _Pbuffer+perow*(SP_BlkSize+2);
    SP_PivotVector[SNma->Sstart[k]+bcol].blockNo = *gbuf;
    SP_PivotVector[SNma->Sstart[k]+bcol].rowNo = *(gbuf+1);

    /* update the block_kk at ppe */
    if (SP_me_no == ppe) {
      px = 1.0/bkk[lda*bcol+bcol];
      n = lda-bcol-1;
      l = 1;
      pa = bkk+lda*bcol+bcol+1;
      spSCAL (&n, &px, pa, &l);      /* scale the first column */
      for (i=bcol+1; i<lda; ++i) {   /* update the rest of the block */
        py = 0.0 - bkk[lda*i+bcol];
        spAXPY (&n, &py, pa, &l, pa+lda*(i-bcol), &l);
      }
    }

    /* updating the rest of the column block */    
    alpha = -1.0;
    n = lda-bcol-1;
    l = 1;
    px=1.0/(gbuf+2)[bcol];
    for (i=stc; i<endc; ++i) { 
      if (SNma->LnodeCol[i]==NULL)
	continue;
      m = SNma->LsubRowNumInBlk[i];
      bik = SNma->LnodeCol[i]+m*bcol;      
      spSCAL (&m, &px, bik, &l);    /* scale */
      spGER (&m, &n, &alpha, bik, &l, gbuf+2+bcol+1, &l, (bik + m), &m);
    }
  }
}

/* returns 1 indicates a success; returns 0 indicates a bailout */
static int _Factor_BatchPivot (SP_SNmatrix_struct *SNma, INDEXTYPE k) {
  MPI_Status SP_status;
  int ppe, perow, pstart;
  int i, l, l2, m, n, lda, bcol, bailout;
  INDEXTYPE stc, endc;
  ELEMTYPE alpha, px, py;
  ELEMTYPE *lbbuf;
  ELEMTYPE *bkk = NULL, *bik = NULL;
  ELEMTYPE *(lpblka[SP_BlkSize]);
  int lpblk[SP_BlkSize], lprow[SP_BlkSize], lprowstorage[SP_BlkSize], lprowtotstorage[SP_BlkSize];

  bailout = 0;

  ppe=PEOWNER(k,k);
  perow=ROWNER(k);
  pstart=COWNER(k)*SP_rp;               /* the pe no of the start of pecol */

  lda = DSIZE (SNma, k);
  stc = SNma->LstartCol[CINDEX(k)];
  endc = SNma->LstartCol[CINDEX(k)+1];
  if (SP_me_no == ppe) bkk = SNma->LnodeCol[stc++];

  /* lbbuf is for holding the local pivoting candidates for the current column block.
     block_no of local maxima, row_no inside the block of local maxima, and elements 
     for the row containing the local maxima. */
  lbbuf = _Pbuffer+SP_me_rno*lda*(lda+2);

  /* find large elements in each local column */
  for (bcol = 0; bcol < lda; ++bcol) {   /* look at each column in the block */
    /* lcbuf contains the local pivoting candidate for the current column */
    ELEMTYPE *lcbuf = lbbuf+bcol*(lda+2);

    px = 0.0;                            /* local maxima */
    lpblka[bcol] = NULL;

    /* bkk is not included in this search */

    for (i = stc; i < endc; ++i) {      /* pivot searching */
      if ((bik = SNma->LnodeCol[i])==NULL)
	continue;

      for (l2 = l = 0; l < DSIZE (SNma, SNma->LindexCol[i]); ++l) {
        if (testcol(SNma->LsubRow[i], l)) {
	  py = fabs (bik[bcol * SNma->LsubRowNumInBlk[i] + l2]);
	  if (py > px) {
	    px = py;
	    lpblk[bcol] = SNma->LindexCol[i];
	    lprow[bcol] = l;
	    lpblka[bcol] = bik;
	    lprowstorage[bcol] = l2;
	    lprowtotstorage[bcol] = SNma->LsubRowNumInBlk[i];
	  }
          l2 ++;
        }
      }
    }

    *lcbuf = lpblk[bcol];
    *(lcbuf+1) = lprow[bcol];

    /* copy the row containing the local maxima to the buffer */
    if (lpblka[bcol]!=NULL && lpblka[bcol][bcol*lprowtotstorage[bcol]+lprowstorage[bcol]]!=0.0) {
      /* find some local maxima */
      for (i=0; i<lda; ++i)
	(lcbuf+2)[i] = lpblka[bcol][i*lprowtotstorage[bcol]+lprowstorage[bcol]];
      /* make sure this row will not be selected for any other column */
      for (i=bcol+1; i<lda; ++i)
	lpblka[bcol][i*lprowtotstorage[bcol]+lprowstorage[bcol]] = 0.0;
    }      
    else {
      *lcbuf = -1;                 /* mark it as invalid */
      for (i=0; i<lda; ++i)
	(lcbuf+2)[i]=0.0;          /* make sure this row will not be selected as pivot by ppe */
    }
  }

  /* recall that some elements in the candidate pivot rows were cleared zero to 
     prevent them from being selected for multiple columns, now we need to 
     revert them back. */
  for (bcol = 0; bcol < lda; ++bcol) {
    /* lcbuf contains the local pivoting candidate for the current column */
    ELEMTYPE *lcbuf = lbbuf+bcol*(lda+2);

    if (*lcbuf>=0) 
      for (i=bcol+1; i<lda; ++i)
        lpblka[bcol][i*lprowtotstorage[bcol]+lprowstorage[bcol]] = (lcbuf+2)[i];
  }

  /* gather all local maxima at ppe (owner of bkk) */
  if (SP_me_no == ppe) {
    for (i=0; i<SP_rp; ++i) 
      if (i != SP_me_rno) {
        MPI_Recv(_Pbuffer+i*lda*(lda+2), lda*(lda+2), ELEMTYPE_MPI, pstart+i, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
      }
  } 
  else {  /* not ppe */
    MPI_Send(lbbuf, lda*(lda+2), ELEMTYPE_MPI, ppe, SP_MPI_TAG, MPI_COMM_WORLD );
  }

  /* determine pivot rows at ppe */
  if(SP_me_no == ppe) {
    ELEMTYPE gmax[SP_BlkSize];
    rowID *pivot = &SP_PivotVector[SNma->Sstart[k]];

    memcpy(_Bkkbuffer, bkk, lda*lda*sizeof(ELEMTYPE));
  
    /* finding the max of each column */ 
    for (bcol = 0; bcol < lda; ++bcol) { /* look at each column in the block */
      gmax[bcol] = 0.0;
      for (l=0; l<lda; ++l) {
	ELEMTYPE _elem = _Bkkbuffer[bcol*lda+l];
	if (fabs(_elem)>gmax[bcol])
          gmax[bcol] = fabs(_elem);
      }	
      for (i=0; i<SP_rp; ++i) {
	for (l=0; l<lda; ++l) {
	  ELEMTYPE _elem = _Pbuffer[(i*lda+l)*(lda+2)+2+bcol];
	  if (fabs(_elem)>gmax[bcol])
	    gmax[bcol] = fabs(_elem);
	}
      }
    }
 
    for (bcol = 0; bcol < lda; ++bcol) { /* look at each column in the block */
      ELEMTYPE *maxrow=NULL, *pb;
      int maxrowinc=0;

      px = 0.0;
      for (l=bcol; l<lda; l++) { /* scan for bkk */
	py = fabs(_Bkkbuffer[bcol*lda+l]) / SP_PivotThresh;
	if (py>px)
	  { px=py; pivot[bcol].blockNo=k; pivot[bcol].rowNo=l; maxrow=_Bkkbuffer+l; maxrowinc=lda; }
      }

      for (i=0; i<SP_rp; ++i) { /* scan for local maximas from all pes */
	for (l=0; l<lda; ++l) {
	  /* candidate pivot row originally chosen for column x at a pe can be
	     eventually used as pivot row for column y (x<y) because it was not 
	     globally maximal for column x but it is better than any candidate
	     pivot row for column y.
	  */
	  ELEMTYPE *thisbuf = _Pbuffer+(i*lda+l)*(lda+2);
	  if (*thisbuf < 0) continue;       /* already selected */

	  py = fabs((thisbuf+2)[bcol]);     /* the local maxima from pe i */
	  if (i == SP_me_rno) py /= SP_PivotThresh;
	                        /* we'd like to find a pivot row in the diagonal
                                   processor to avoid row swapping, as long as 
                                   the maximum candidate pivot in the diagonal 
                                   processor is not too small compared with the 
                                   global maximum pivot.
                                   we use threshold pivoting here */
	  if (py>px || (py==px && *thisbuf<pivot[bcol].blockNo))
	    { px=py; pivot[bcol].blockNo=*thisbuf; pivot[bcol].rowNo=*(thisbuf+1); maxrow=thisbuf+2; maxrowinc=1; }
	}	
      }

      if (px<SP_Epsilon || px<SP_EpsilonBPratio*gmax[bcol]) {
	bailout = 1;
	break;
      }
      for (l=0; l<lda; l++) { /* interchange rows */
	py = _Bkkbuffer[l*lda+bcol];
	_Bkkbuffer[l*lda+bcol] = maxrow[l*maxrowinc];
	maxrow[l*maxrowinc] = py;
      }

      /* batch pivot with speculative factorization */
      if (SP_BatchPivot==SP_PIVOTING_SBP) {
	/* factorizing bkk */
        px = 1.0/_Bkkbuffer[bcol*lda+bcol];
	n = lda-bcol-1;
	l = 1;
	pb = _Bkkbuffer+bcol*lda+bcol+1;
	spSCAL (&n, &px, pb, &l);	/* scale the first column */
	for (i=bcol+1; i<lda; ++i) {    /* update the rest of the block */
	  py = 0.0 - _Bkkbuffer[i*lda+bcol];
	  spAXPY (&n, &py, pb, &l, pb+(i-bcol)*lda, &l);
	}

	/* factorizing the rest */
        n = lda*SP_rp;
        l = lda+2;
        pb = _Pbuffer+2+bcol;
        spSCAL (&n, &px, pb, &l);	/* scale the first column */
	m = lda-bcol-1;
	alpha = -1.0;
	spGER (&m, &n, &alpha, _Bkkbuffer+lda*(bcol+1)+bcol, &lda, pb, &l, pb+1, &l);
      }
    }

    /* batch pivot with large diagonals */
    if (SP_BatchPivot==SP_PIVOTING_LD && !bailout) {
      /* tentative update the block_kk at ppe.  since we are not sure if this will 
	 succeed, we do things in a buffer (_Bkkbuffer) space, instead of the bkk itself */
      for (bcol = 0; bcol < lda; ++bcol) {
	ELEMTYPE *pb;
	px = _Bkkbuffer[lda*bcol+bcol];
	if (fabs(px)<SP_Epsilon || fabs(px)<SP_EpsilonBPratio*gmax[bcol]) { 
	  bailout = 1;
	  break;
	}
	px = 1.0/px;
	n = lda-bcol-1;
	l = 1;
	pb = _Bkkbuffer+lda*bcol+bcol+1;
	spSCAL (&n, &px, pb, &l);      /* scale the first column */
	for (i=bcol+1; i<lda; ++i) {   /* update the rest of the block */
	  py = 0.0 - _Bkkbuffer[lda*i+bcol];
	  spAXPY (&n, &py, pb, &l, pb+lda*(i-bcol), &l);
	}
      }
    }

    if (!bailout) {
      for (i=0; i<SP_rp; ++i) _SwapCounts[i] = 0;

      /* get ready to the pivot row broadcast and tentative swaps */
      for (bcol = 0; bcol < lda; ++bcol) {   /* look at each column in the block */
	int rpno;

	/* no exchange */
	if (pivot[bcol].blockNo==k && pivot[bcol].rowNo==bcol) continue;

	/* exchange within bkk */
	if (pivot[bcol].blockNo==k) {
	  for (i=0; i<lda; ++i) bkk[lda*i+pivot[bcol].rowNo] = bkk[lda*i+bcol];
	  continue;
	}
	
	rpno = ROWNER(pivot[bcol].blockNo);
	if (rpno == SP_me_rno) { /* global maxima is in ppe - local swap */
	  int _bcol;
	  /* recall that the pivot row may be the candidate row for another column */
	  for (_bcol = 0; _bcol < lda; ++_bcol) {
	    if (lpblka[_bcol]!=NULL && pivot[bcol].blockNo==lpblk[_bcol] && pivot[bcol].rowNo==lprow[_bcol]) {
	      for (i=0; i<lda; ++i) 
		lpblka[_bcol][lprowtotstorage[_bcol]*i+lprowstorage[_bcol]] = bkk[lda*i+bcol];
	    }
	  }
	}
	else { /* global maxima is not in ppe - update the buf row (part of remote swap) */
	  ELEMTYPE *_buf = _Pbuffer+(rpno*lda+_SwapCounts[rpno])*(lda+2);
	  *_buf = pivot[bcol].blockNo;
	  *(_buf+1) = pivot[bcol].rowNo;
	  for (i=0; i<lda; ++i) (_buf+2)[i] = bkk[lda*i+bcol];
	  _SwapCounts[rpno]++;
	}
      }

      memcpy(bkk, _Bkkbuffer, lda*lda*sizeof(ELEMTYPE));
    }
  }

  if (SP_me_no == ppe) {
    if (bailout)  {
      for (i=0; i<SP_rp; ++i)
	if (i != SP_me_rno) {
	  /* bailout: send a zero-sized message */
          MPI_Send(&SP_PivotVector[SNma->Sstart[k]], 0, MPI_CHAR, pstart+i, SP_MPI_TAG, MPI_COMM_WORLD);
	}
      return 0;
    }
    else {
      for (i=0; i<SP_rp; ++i)
	if (i != SP_me_rno) {
	  /* broadcast the pivot row from ppe (owner of bkk) */
          MPI_Send(&SP_PivotVector[SNma->Sstart[k]], lda*sizeof(rowID), MPI_CHAR, pstart+i, SP_MPI_TAG, MPI_COMM_WORLD);

	  /* broadcast the updated pivot row */
	  MPI_Send(bkk, lda*lda, ELEMTYPE_MPI, pstart+i, SP_MPI_TAG, MPI_COMM_WORLD);

	  /* broadcast the pivot row or the swapped row */
	  MPI_Send(_Pbuffer+i*lda*(lda+2), _SwapCounts[i]*(lda+2), ELEMTYPE_MPI, pstart+i, SP_MPI_TAG, MPI_COMM_WORLD);
	}
    }
  }
  else {  /* not ppe */
    int count;
    MPI_Recv(&SP_PivotVector[SNma->Sstart[k]], lda*sizeof(rowID), MPI_CHAR, ppe, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
    MPI_Get_count(&SP_status, ELEMTYPE_MPI, &count);
    if (count==0) return 0; /* bail out */

    bkk = _Bkkbuffer;
    MPI_Recv(bkk, lda*lda, ELEMTYPE_MPI, ppe, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
    MPI_Recv(_Pbuffer+perow*lda*(lda+2), lda*(lda+2), ELEMTYPE_MPI, ppe, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
    MPI_Get_count(&SP_status, ELEMTYPE_MPI, &count);
    count = count / (lda+2);

    for (l = 0; l < count; l++) {
      ELEMTYPE *_buf = _Pbuffer+(perow*lda+l)*(lda+2);
      int _bcol;
      for (_bcol = 0; _bcol < lda; ++_bcol) {
	if (lpblka[_bcol] !=NULL && *_buf == lpblk[_bcol] && *(_buf+1) == lprow[_bcol]) { /* this pe has the maximum element */
	  for (i=0; i<lda; ++i) { /* interchange row with the buf row */
	    py = lpblka[_bcol][lprowtotstorage[_bcol]*i+lprowstorage[_bcol]]; 
	    lpblka[_bcol][lprowtotstorage[_bcol]*i+lprowstorage[_bcol]] = (_buf+2)[i];
	    (_buf+2)[i] = py; 
	  }  
	}
      }
    }
  }

  for (bcol = 0; bcol < lda; ++bcol) {      
    /* updating the rest of the column block */    
    alpha = -1.0;
    n = lda-bcol-1;
    l = 1;
    px=1.0/bkk[lda*bcol+bcol];
    for (i=stc; i<endc; ++i) { 
      if (SNma->LnodeCol[i]==NULL)
	continue;
      m = SNma->LsubRowNumInBlk[i];
      bik = SNma->LnodeCol[i]+m*bcol;      
      spSCAL (&m, &px, bik, &l);    /* scale */
      spGER (&m, &n, &alpha, bik, &l, bkk+lda*(bcol+1)+bcol, &lda, (bik + m), &m);
    }
  }

  return 1;
}

static void _Factor (SP_SNmatrix_struct *SNma, INDEXTYPE k) {
  if (SP_BatchPivot != SP_PIVOTING_COLBYCOL) {
    if (!_Factor_BatchPivot(SNma, k)) {
      SP_BatchPivotBailout++;
      _Factor_SerialPivot(SNma, k);
    }
  }
  else
    _Factor_SerialPivot(SNma, k);

  /* send column, pivot vector and L^-1 */
  if (k<SNma->snOrder) {
    SP_send_pivot_MPI(SNma, k);
    if (SP_me_no==PEOWNER(k,k)) SP_send_linv_MPI(SNma, k);
    SP_send_col_MPI(SNma, k);
  }
}

static void _Scale (SP_SNmatrix_struct *SNma, INDEXTYPE k) {
  MPI_Status SP_status;
  register INDEXTYPE i, str, endr;
  ELEMTYPE alpha = 1.0;
  int n, m;
  ELEMTYPE *Ublk;

  /* get L inverse from PEOWNER(k), used to be SP_get_linv_MPI */
  MPI_Recv(_Bkkbuffer, DSIZE(SNma, k)*DSIZE(SNma, k), ELEMTYPE_MPI, PEOWNER(k, k),
	   SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);

  n = DSIZE (SNma, k);
  str = SNma->UstartRow[RINDEX(k)];
  endr = SNma->UstartRow[RINDEX(k) + 1];
  for (i = str; i < endr; i++) {
      if(!(Ublk = SNma->UnodeRow[i]))	continue;

      m = SNma->UsubColNumInBlk[i];
      spTRMM (SP_ftcL, SP_ftcL, SP_ftcN, SP_ftcU, &n, &m, &alpha, _Bkkbuffer, &n, Ublk, &n);
  }

  SP_send_row_MPI(SNma, k);
}

#define _Modify_Uij(SNma, blk_ik, blk_kj, blk_ij, i, j, k, rik, ckj, cij) \
	SP_Sdgemmc(DSIZE(SNma, i), DSIZE(SNma, j), DSIZE(SNma, k), blk_ik, SP_OrderOfInt(rik), rik, \
	blk_kj, SP_OrderOfInt(ckj), ckj, blk_ij, SP_OrderOfInt(cij), cij)

#define _Modify_Lij(SNma, blk_ik, blk_kj, blk_ij, i, j, k, rik, ckj, rij) \
	SP_Sdgemmr(DSIZE(SNma, i), DSIZE(SNma, j), DSIZE(SNma, k), blk_ik, SP_OrderOfInt(rik), rik, \
        blk_kj, SP_OrderOfInt(ckj), ckj, blk_ij, SP_OrderOfInt(rij), rij)

static void _Update (SP_SNmatrix_struct *SNma, INDEXTYPE k, INDEXTYPE from, INDEXTYPE to) {
  register int i, j;
  register INDEXTYPE tr, tc;
  static INDEXTYPE stc, endc, str, endr, tor;
  register INDEXTYPE lstri, lendri, ustcj, uendcj;
  PATTYPE *ckj, *cij, *rik, *rij;
  PATTYPE *rikStart = NULL, *ckjStart = NULL;
  ELEMTYPE *ukj, *uij, *lik = NULL, *lij;
  ELEMTYPE *likStart = NULL, *ukjStart = NULL;
  int rpe, cpe;
  INDEXTYPE *colindbuf = NULL, *rowindbuf;

  rpe=ROWNER(k);
  cpe=COWNER(k);

  if (cpe == SP_me_cno) {
    stc = SNma->LstartCol[CINDEX(k)];
    if (SNma->LindexCol[stc] == k) stc++;
    endc = SNma->LstartCol[CINDEX(k)+1];
	if (stc<endc) {
      likStart = SNma->LnodeCol[stc];
      rikStart = SNma->LsubRow+stc;
      colindbuf = SNma->LindexCol;
	}
  }
  else {
    stc=0;
    likStart = SP_Cbuffer;
    rikStart = SP_Cpatternbuf;
    colindbuf = SP_Cindexbuf;
    if (to) endc=SP_get_col_MPI(SNma, k);
  }

  if (rpe == SP_me_rno) {
    str = SNma->UstartRow[RINDEX(k)];
    endr = SNma->UstartRow[RINDEX(k)+1];
    rowindbuf = SNma->UindexRow;
    while (str<endr && rowindbuf[str]<from) str++;
	if (str<endr) {
	  ukjStart = SNma->UnodeRow[str];
	  ckjStart = SNma->UsubCol+str;
	}
  }
  else {
    str = 0;
    rowindbuf = SP_Rindexbuf;
    ukjStart = SP_Rbuffer;
    ckjStart = SP_Rpatternbuf;
    if (to) endr=SP_get_row_MPI(SNma, k);
    while (str<endr && rowindbuf[str]<from) {
      str++; 
      ukjStart += SP_OrderOfInt(*ckjStart)*DSIZE(SNma, k); 
      ckjStart++;
    }
  }

  if (from>SNma->snOrder) return;

  if (str>=endr || stc>=endc) return;

  if (to==1) {
    if (rowindbuf[str]>from) return;
    tor=str+1;
  }
  else tor=endr;

  ukj=ukjStart;
  ckj=ckjStart;
  for (tr=str; tr<tor; tr++, ckj++) {
    /* if I'm a row owner and the block is null, or if Cpatternbuf2 says it's null, continue */
    if ((rpe==SP_me_rno && (ukj=SNma->UnodeRow[tr])==NULL) || !*ckj)
      continue;
    j = rowindbuf[tr];

    /* get the index of L of j-th column */
    lstri = SNma->LstartCol[CINDEX(j)];
    lendri = SNma->LstartCol[CINDEX(j)+1];

    /* if no blocks, do not wait for Rbuffer */
    if (lstri<lendri) {

      tc = stc;
      rik = rikStart;
      if (cpe==SP_me_cno)
	for (; tc<endc && colindbuf[tc]<j; ++tc, ++rik);
      else {
	lik = likStart;
	while (tc<endc && colindbuf[tc]<j) {
	  tc++;
	  lik+=SP_OrderOfInt(*rik)*DSIZE(SNma, k);
	  rik++;
	}
      }
      
      for (; tc<endc; tc++, rik++) {
	if ((cpe==SP_me_cno && (lik=SNma->LnodeCol[tc])==NULL) || !*rik)
	  continue;

	i = colindbuf[tc];
	
	while (lstri<lendri && SNma->LindexCol[lstri]<i) lstri++;
	if (SNma->LindexCol[lstri]==i) {
	  if (!SNma->LnodeCol[lstri])
	    SNma->LnodeCol[lstri] = SP_MallocElem( SNma->LsubRowNumInBlk[lstri]*DSIZE(SNma, j) );
	  lij = SNma->LnodeCol[lstri];
	  rij = SNma->LsubRow + lstri;
	  _Modify_Lij(SNma, lik, ukj, lij, i, j, k, *rik, *ckj, *rij);
	}
	if (cpe != SP_me_cno)
	  lik+=SP_OrderOfInt(*rik)*DSIZE(SNma, k);
      }
    }
    if (rpe != SP_me_rno)
      ukj+=SP_OrderOfInt(*ckj)*DSIZE(SNma, k);
  }

  lik = likStart;
  rik = rikStart;
  for (tc = stc; tc < endc; tc++, rik++) {
    if ((cpe==SP_me_cno && (lik=SNma->LnodeCol[tc])==NULL) || !*rik)
      continue;

    i = colindbuf[tc];

    /* get the index of U of j-th row */
    ustcj = SNma->UstartRow[RINDEX(i)];
    uendcj = SNma->UstartRow[RINDEX(i)+1];

    if (ustcj<uendcj) {
      tr = str;
      ckj = ckjStart;
      if (rpe==SP_me_rno)
	for (;tr<tor && rowindbuf[tr]<=i; ++tr, ++ckj);
      else {
	ukj = ukjStart;
	while (tr<tor && rowindbuf[tr]<=i) {
	  tr++;
	  ukj+=SP_OrderOfInt(*ckj)*DSIZE(SNma, k);
	  ckj++;
	}
      }

      for (; tr<tor; tr++, ckj++) {
	if ((rpe==SP_me_rno && (ukj=SNma->UnodeRow[tr])==NULL) || !*ckj)
	  continue;
	j = rowindbuf[tr];
	
	while (ustcj<uendcj && SNma->UindexRow[ustcj]<j) ustcj++;
	if (SNma->UindexRow[ustcj]==j) {
	  if (!SNma->UnodeRow[ustcj])
	    SNma->UnodeRow[ustcj] = SP_MallocElem( SNma->UsubColNumInBlk[ustcj]*DSIZE(SNma, i) );

	  uij = SNma->UnodeRow[ustcj];
	  cij = SNma->UsubCol + ustcj;
	  _Modify_Uij(SNma, lik, ukj, uij, i, j, k, *rik, *ckj, *cij);
	}
	if (rpe != SP_me_rno)
	  ukj+=SP_OrderOfInt(*ckj)*DSIZE(SNma, k);
      }
    }
    if (cpe != SP_me_cno)
      lik+=SP_OrderOfInt(*rik)*DSIZE(SNma, k);
  }
}

/* LU factorization with partial pivoting */
void SP_LUP (SP_SNmatrix_struct *SNma, rowID **Prow) {
  register INDEXTYPE k, k1;
  short *factored;

  factored = (short*) SP_Malloc((SNma->snOrder/SP_cp+1)*sizeof(short));
  bzero(factored, (SNma->snOrder/SP_cp+1)*sizeof(short));

  _InitLUP(SNma);

  if (SP_me_cno == 0) _Factor (SNma, 1);

  for (k = 1; k < SNma->snOrder; k++) {
    k1 = CINDEX(k)*SP_cp+SP_me_cno+1;
    if (k1 <= k) k1+=SP_cp;

    SP_Interchange (SNma, k);

    if (SP_me_rno == ROWNER(k)) _Scale (SNma, k);

    if (k1<=SNma->snOrder && !factored[CINDEX(k1)] && k>=SNma->SnodeChild[CINDEX(k1)]) {
      _Update (SNma, k, k1, 1);
      _Factor (SNma, k1);
      factored[CINDEX(k1)] = 1;
      _Update (SNma, k, k1+SP_cp, 0);
    }
    else {
      _Update (SNma, k, k1, 2);
    }
  }

  SP_Free (factored);

  _FinalLUP ();

  *Prow = SP_PivotVector;
}
