/********************************************************************* 
-- S+ sparse GE solver (version 1.1) --

Copyright (C) 1999 by University of California at Santa Barbara. 
Copyright (C) 2005 by University of Rochester. 
All rights reserved.

S+ team contact: Kai Shen <kshen@cs.rochester.edu>, 
                 Tao Yang <tyang@cs.ucsb.edu>  

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose, without fee, and without written agreement
is hereby granted, provided that the above copyright notice and the
following paragraph appear in all copies of this software.

This software is provided ``as is'', without any warranty, including 
any implied warranty for merchantability or fitness for a particular
purpose. Under no circumstances shall the University of California at 
Santa Barbara, the University of Rochester, or their agents be liable 
for any use of, misuse of, or inability to use this software, including 
incidental and consequential damages.
************************************************************************/

/*
 * Sblas.c - supernodal BLAS routines
 *
 */


#include <stdio.h>
#include <math.h>

#include "struct.h"
#include "functions.h"
#include "globals.h"

static ELEMTYPE *_TmpBlk;

void SP_InitSblas() {
  _TmpBlk = (ELEMTYPE*) SP_Malloc(SP_BlkSize*SP_BlkSize*sizeof(ELEMTYPE));
}

void SP_FinalSblas() {
  SP_Free(_TmpBlk);
}

static int _StructDiff(int sub, int sup) {
  register int i=1;
  register int diff;

  diff = 0;
  while (sup)
    {
      if (sup & 1)
	{
	  if (sub & 1)
	    diff |= i;
	  i <<= 1;
	}
      else if (sub & 1)
	SP_err(SP_ERR_LOC, "sub is not a subset of sup: sub=0x%X sup=0x%X", sub, sup);
      sup >>= 1;
      sub >>= 1;
    }
  return diff;
}

void SP_Sdgemmc(int M, int N, int K, 
        ELEMTYPE *A, int nrA, int rowA, 
        ELEMTYPE *B, int ncB, int colB, 
        ELEMTYPE *C, int ncC, int colC)
{
  register int    i, r, c;
  register int   col;
  ELEMTYPE *p;
  register ELEMTYPE *pC=C;
  ELEMTYPE alpha = -1.0, beta=1.0, zero=0.0;
  int       one = 1;

  if (!rowA || !colB)
    SP_err(SP_ERR_LOC, "ERROR in Sdgemmc");
  if ( M == nrA && colB == colC) {
    spGEMM (SP_ftcN,SP_ftcN,&nrA,&ncB,&K,&alpha,A,&nrA,B,&K,&beta,C,&nrA);
  }
  else
    {
      spGEMM (SP_ftcN,SP_ftcN,&nrA,&ncB,&K,&alpha,A,&nrA,B,&K,&zero,_TmpBlk,&nrA);

      col =  _StructDiff(colB, colC);
      p = _TmpBlk;
      if (nrA == M)
        {
          for (c = 0; c < ncC; c++,pC+=M)
            {
              if (!testcol(col, c))
                continue;
              spAXPY(&M, &beta, p, &one, pC, &one);
              p+=nrA;
            }
        }
     else if (colB == colC)
        {
          for (r = 0; r < M; r++, pC++)
            {
              if (!testcol(rowA, r))
                continue;
              spAXPY(&ncC, &beta, p, &nrA, pC, &M);
              p++;
            }
        }
      else
        {
          for (c = 0; c < ncC; c++, pC+=M)
	    {
	      if (!testcol(col, c))
	        continue;
	      for (r = 0,i=0; r < M; r++)
	        if (testcol(rowA, r)) 
	          pC[r] += p[i++];
              p+=nrA;
	    }
        }
    }
}

void SP_Sdgemmr(int M, int N, int K, 
        ELEMTYPE *A, int nrA, int rowA, 
        ELEMTYPE *B, int ncB, int colB, 
        ELEMTYPE *C, int nrC, int rowC)
{
  register int i, r, c;
  register int row;
  ELEMTYPE *p;
  register ELEMTYPE *pC=C;
  ELEMTYPE alpha = -1.0, beta=1.0, zero=0.0;
  int    one = 1;

  if (!rowA || !colB)
    SP_err(SP_ERR_LOC, "ERROR in Sdgemmr");
  if ( N == ncB && rowA == rowC) {
    spGEMM (SP_ftcN,SP_ftcN,&nrA,&ncB,&K,&alpha,A,&nrA,B,&K,&beta,C,&nrA);
  }
  else
    {
      spGEMM (SP_ftcN,SP_ftcN,&nrA,&ncB,&K,&alpha,A,&nrA,B,&K,&zero,_TmpBlk,&nrA);

      p = _TmpBlk;

      if (rowA == rowC)
        {
          for (c = 0; c < N; c++, pC+=nrC)
            {
              if (!testcol(colB, c))
                continue;
              spAXPY(&nrC, &beta, p, &one, pC, &one);
              p+=nrA;
            }
        }
      else if (ncB == N)
        {
          row = _StructDiff(rowA, rowC);
          for (r = 0; r < nrC; r++, pC ++)
            {
              if (!testcol(row, r))
                continue;
              spAXPY(&N, &beta, p, &nrA, pC, &nrC);
              p++;
            }
        }
      else
        {
          row = _StructDiff(rowA, rowC);
          for (c = 0; c < N; c++, pC+=nrC)
            {
              if (!testcol(colB, c))
                continue;
              for (r = 0,i=0; r < nrC; r++)
                if (testcol(row, r)) 
                  pC[r] += p[i++];
              p+=nrA;
            }
        }
    }
}

