/********************************************************************* 
-- S+ sparse GE solver (version 1.1) --

Copyright (C) 1999 by University of California at Santa Barbara. 
Copyright (C) 2005 by University of Rochester. 
All rights reserved.

S+ team contact: Kai Shen <kshen@cs.rochester.edu>, 
                 Tao Yang <tyang@cs.ucsb.edu>  

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose, without fee, and without written agreement
is hereby granted, provided that the above copyright notice and the
following paragraph appear in all copies of this software.

This software is provided ``as is'', without any warranty, including 
any implied warranty for merchantability or fitness for a particular
purpose. Under no circumstances shall the University of California at 
Santa Barbara, the University of Rochester, or their agents be liable 
for any use of, misuse of, or inability to use this software, including 
incidental and consequential damages.
************************************************************************/

/*
 * SymFac.c - symbolic factorization routine
 *
 */


#include <stdio.h>
#include <string.h>

#include "struct.h"
#include "functions.h"
#include "globals.h"

static NodeList *_NLhead = NULL;
static NodeList *_NLtail = NULL;

/* memory allocation for NodeList */
static NodeList *_MallocNL() {
  static short size = sizeof(NodeList);
  NodeList *tmp;

  if (_NLhead != NULL) {
    tmp = _NLhead;
    _NLhead = _NLhead->next;
    if (_NLhead==NULL) _NLtail=NULL;
    return tmp;
  }
  else
    return (NodeList*)SP_Malloc(size);
} 

/* memory free for NodeList */
void SP_FreeNL(NodeList *freed) {
  freed->next = NULL;
  if (_NLtail == NULL) 
    _NLhead = _NLtail = freed;
  else { 
    _NLtail->next = freed;
    _NLtail = freed;
  }
} 

/* free all NodeList */
void SP_FreeAllNL() {
  register NodeList *tmp;

  while (_NLhead != NULL) {
    tmp = _NLhead;
    _NLhead = _NLhead->next;
    SP_FreeSize(tmp, sizeof(NodeList));
  }
  _NLtail = NULL;
}

/* free the space of the whole matrix */
void SP_FreeMatrix(Matrix *ma) {
  register INDEXTYPE i;
  register NodeList *tmp;

  SP_FreeSize(ma->colSize, (ma->order+1)*sizeof(INDEXTYPE));
  SP_FreeSize(ma->rowSize, (ma->order+1)*sizeof(INDEXTYPE));
  for (i=1; i<=ma->order; i++) {
    while (ma->col[i]!=NULL && ma->col[i]->from!=-1) {
      tmp = ma->col[i];
      ma->col[i] = ma->col[i]->next;
      tmp->from=-1;
      SP_FreeNL(tmp);
    }
    while (ma->row[i]!=NULL && ma->row[i]->from!=-1) {
      tmp = ma->row[i];
      ma->row[i] = ma->row[i]->next;
      tmp->from=-1;
      SP_FreeNL(tmp);
    }
  }
  SP_FreeSize(ma->col, (ma->order+1)*sizeof(NodeList*));
  SP_FreeSize(ma->row, (ma->order+1)*sizeof(NodeList*));
  SP_FreeSize(ma, sizeof(Matrix));

  SP_FreeAllNL();
}

/* free elimination forest */
void SP_FreeEforest(Node **ef, INDEXTYPE order) {
  INDEXTYPE i;
  Node *tmp;

  for (i=1; i<=order; i++) 
    while (ef[i]!=NULL) {
      tmp = ef[i];
      ef[i] = ef[i]->next;
      SP_FreeSize(tmp, sizeof(Node));
    }

  SP_FreeSize(ef, (order+1)*sizeof(Node*));
}

/*******************************************************
 *  function:     _DeleteNodes()
 *  description:  delete some redundant node from level-1 index structure
 *  return:       number of redundant entries from nextentry->from to 
 *                the last contigious entry
 *******************************************************/
static INDEXTYPE _DeleteNodes (
     NodeList **nextentry, /*In/Out*/
     INDEXTYPE *lastrow,   /*In/Out*/
     INDEXTYPE *allocated  /*In/Out*/
) {
  INDEXTYPE   tto=0, tdel=0;
  NodeList    *p;

  while (*nextentry && (*nextentry)->from-1 <= *lastrow )
    {
      p = *nextentry;
      tto = p->to;
      *nextentry = p->next;
      tdel += tto - p->from +1;
      SP_FreeNL(p);
      (*allocated)--;
    }
  if (tto > *lastrow)
    *lastrow = tto;
  return tdel;
}


/*******************************************************
 *  function:     SP_InsertToColumn()
 *  description:  insert an entry into level-1 compressed column index
 *                this function assume the column entries are
 *                column major manner sorted for best performance
 *                correctness holds if not sorted properly
 *  return:       number of new entries
 *  arguments:    a set of contiguous entries
 *******************************************************/
INDEXTYPE SP_InsertToColumn(NodeList **column, INDEXTYPE from, INDEXTYPE to) {
  NodeList  *newentry;
  static NodeList  *lastcol;
  static NodeList  *lastentry=NULL, *nextentry=NULL;  
  register INDEXTYPE    temp;
  INDEXTYPE    tto, tdel;
  static INDEXTYPE allocated=0;

  if (!column)
    {
      temp = allocated;
      allocated = 0;
      return temp;
    }
  /* Case 1: Add a new entry to the beginning of the column */
  /* Case 1.1: first node in the column*/
  if ( ! *column )
    {
      lastentry = lastcol = *column = _MallocNL();
      lastcol->from = from;
      lastcol->to = to;
      lastcol->next = NULL;
      
      allocated++;
      return to-from+1;
    }
  /* Case 1.2: add a new node right before the first entry of the column
     here the new node is phsically added after the first entry to make
     the memory space contigious after level-2 compression later */
  else if ( to < column[0]->from-1)
    {
      lastcol = *column;
      lastentry = newentry = _MallocNL();
      memcpy(newentry, lastcol, sizeof(NodeList));
      lastcol->from = from;
      lastcol->to = to;
      lastcol->next = newentry;
      
      allocated++;
      return to-from+1;
      
    }
  /* Case 2: Append to the first node */
  else if ( from < column[0]->from)
    {
      lastentry = lastcol = column[0];
      nextentry = lastentry->next;
      temp = lastentry->from;
      lastentry->from = from;
      if (to > lastentry->to)
	{
	  temp = lastentry->to-temp;
	  tto = to;
	  tdel = _DeleteNodes(&nextentry, &tto, &allocated);
	  lastentry->to = tto;
	  lastentry->next = nextentry;
	  return tto-from-tdel-temp;
	}
      else
	return temp - from;
    }
  else
    {
      /* find out where to be inserted */
      if (lastcol != column[0] || lastentry->from > from )
	lastentry = lastcol = column[0];
      
      nextentry = lastentry->next;
      while (nextentry && from >= nextentry->from )
	{
	  lastentry = nextentry;
	  nextentry = lastentry->next;
	}
      
      /* Case 3: redundant entries */
      if (to <= (temp=lastentry->to))
	return 0;
      /* Case 4: Append to the end of lastentry */
      else if ( from <= temp+1)
	{
	  tto = to;
	  tdel = _DeleteNodes(&nextentry, &tto, &allocated);
	  lastentry->to = tto;
	  lastentry->next = nextentry;
	  return tto-temp-tdel;
	}
      /* Case 5: Append to the beginning of nextentry */
      else if (nextentry && to >= nextentry->from-1)
	{
	  lastentry = nextentry;
	  nextentry = nextentry->next;
	  temp = lastentry->from;
	  lastentry->from = from;
	  if (lastentry->to < to)
	    {
	      temp = lastentry->to - temp;
	      tto = to;
	      tdel = _DeleteNodes(&nextentry, &tto, &allocated);
	      lastentry->to = tto;
	      lastentry->next = nextentry;
	      
	      return tto-from-temp-tdel;
	    }
	  else
	    return temp-from;
	}
      /* Case 6: Add a new node */
      else
	{	  
	  newentry = _MallocNL();
	  newentry->from = from;
	  newentry->to = to;
	  newentry->next = nextentry;
	  lastentry->next = newentry;
	  /* prepare for next call */
	  lastentry = newentry;
	  allocated++;
	  return to-from+1;
	}
    }      
}

static Matrix *_FormatMatrix(SP_SparseMatrix *input, int *perm_c, int *perm_r) {
  Matrix *output;
  INDEXTYPE i, col, row, curCol;
  register INDEXTYPE order = input->order;

  output = (Matrix*) SP_Malloc(sizeof(Matrix));
  output->order = order;
  output->col = (NodeList**) SP_Malloc((order+1) * sizeof(NodeList*));
  output->colSize = (INDEXTYPE*) SP_Malloc((order+1) * sizeof(INDEXTYPE));
  output->row = (NodeList**) SP_Malloc((order+1) * sizeof(NodeList*));
  output->rowSize = (INDEXTYPE*) SP_Malloc((order+1) * sizeof(INDEXTYPE));
  
  for (i=1; i<=order; i++) {
    output->col[i] = _MallocNL();
    output->col[i]->from = output->col[i]->to = i;
    output->col[i]->next = NULL;
    output->colSize[i] = 1;
    output->row[i] = _MallocNL();
    output->row[i]->from = output->row[i]->to = i;
    output->row[i]->next = NULL;
    output->rowSize[i] = 1;
  }

  curCol = 1;
  for (i=0; i<input->numNZ; i++) {
    row = input->rowIndex[i]+1;

    if (i == input->colStart[curCol]) curCol++;
    col = curCol;

    /* permute row and col */
    col = perm_c[col-1]+1;
    row = perm_r[row-1]+1;

    if (row >= col) {
      output->colSize[col] += SP_InsertToColumn(output->col+col, row, row);
    }
    output->rowSize[row] += SP_InsertToColumn(output->row+row, col, col);
  }

  return output;
}

/* insert one list into another */
static INDEXTYPE _InsertList(NodeList *tmp1, NodeList **tmp2) {
  INDEXTYPE total=0;

  while (tmp1 != NULL) {
    total += SP_InsertToColumn(tmp2, tmp1->from, tmp1->to);
    tmp1 = tmp1->next;
  }    

  return total;
} 

/* free the whole list of NodeList */
static void _ReleaseList(NodeList *nl) {
  register NodeList *tmp;

  while (nl != NULL) {
    tmp = nl;
    nl = nl->next;
    SP_FreeNL(tmp);
  }
}

/* symbolic factorization */
Matrix *SP_SymFac(SP_SparseMatrix *input, Node **eforest, int *perm_c, int *perm_r) {
  INDEXTYPE i, k, index;
  Matrix *internal, *output;
  NodeList *tmp1;
  Node *tmp2;
  char *mask;
  short copy;
  register INDEXTYPE order=input->order;

  internal = _FormatMatrix(input, perm_c, perm_r);

  output = (Matrix*) SP_Malloc(sizeof(Matrix));
  output->col = (NodeList**) SP_Malloc((order+1) * sizeof(NodeList*));
  output->colSize = (INDEXTYPE*) SP_Malloc((order+1) * sizeof(INDEXTYPE));
  output->row = (NodeList**) SP_Malloc((order+1) * sizeof(NodeList*));
  output->rowSize = (INDEXTYPE*) SP_Malloc((order+1) * sizeof(INDEXTYPE));
  output->order = order;

  mask = (char*) SP_Malloc((order+1) * sizeof(char));
  for (k=1; k<=order; k++) mask[k]=0;

  for (k=1; k<=order; k++) eforest[k] = NULL;

  for (k=1; k<=order; k++) {
    output->col[k] = NULL;
    output->colSize[k] = 0;
    output->row[k] = NULL;
    output->rowSize[k] = 0;
    copy = 1;

    tmp1 = internal->col[k];
    while (tmp1 != NULL) {
      for (i=tmp1->from; i<=tmp1->to; i++) {
        if (mask[i] == 0) {
	  output->colSize[k] += SP_InsertToColumn(output->col+k, i, i);
	  output->rowSize[k] += _InsertList(internal->row[i], &(output->row[k]));
	  _ReleaseList(internal->row[i]); internal->row[i]=NULL;
	  mask[i] = 1;
	  copy = 0;
        }
      }
      tmp1 = tmp1->next;
    }
    _ReleaseList(internal->col[k]); internal->col[k]=NULL;

    if (copy && (eforest[k] != NULL) && (eforest[k]->next == NULL)) {
      index = eforest[k]->index;
      output->colSize[k] = output->colSize[index]-1;
      output->rowSize[k] = output->rowSize[index]-1;
      if (index != k-1) {
	output->col[k] = output->col[index]->next;
	output->row[k] = output->row[index]->next;
      }
      else {
	output->col[k] = _MallocNL();
	output->col[k]->from = k;
	output->col[k]->to = output->col[index]->to;
	output->col[k]->next = output->col[index]->next;
	output->row[k] = _MallocNL();
	output->row[k]->from = k;
	output->row[k]->to = output->row[index]->to;
	output->row[k]->next = output->row[index]->next;
      }
    }
    else {
      tmp2 = eforest[k];
      while (tmp2 != NULL) {
	i = tmp2->index;
      
	tmp1 = output->col[i];
	while (tmp1 != NULL) {
	  if (tmp1->to >= k) {
	    if (tmp1->from >= k)
	      output->colSize[k] += SP_InsertToColumn(output->col+k, tmp1->from, tmp1->to);
	    else
	      output->colSize[k] += SP_InsertToColumn(output->col+k, k, tmp1->to);
	  }
	  tmp1 = tmp1->next;
	}

	tmp1 = output->row[i];
	while (tmp1 != NULL) {
	  if (tmp1->to >= k) {
	    if (tmp1->from >= k)
	      output->rowSize[k] += SP_InsertToColumn(output->row+k, tmp1->from, tmp1->to);
	    else
	      output->rowSize[k] += SP_InsertToColumn(output->row+k, k, tmp1->to);
	  }
	  tmp1 = tmp1->next;
	}

	tmp2 = tmp2->next;
      }
    } 

    if (output->colSize[k] > 1) {
      tmp1 = output->row[k];
      i = -1;
      while (tmp1 != NULL) {
	if (tmp1->to > k) {
	  if (tmp1->from > k) i=tmp1->from;
	  else i = k+1;
	  break;
	}
	tmp1 = tmp1->next;
      }
      if (tmp1 != NULL) {
	tmp2 = (Node*) SP_Malloc(sizeof(Node));
	tmp2->index = k;
	tmp2->next = eforest[i];
	eforest[i] = tmp2;
      }
    }
  }

  SP_FreeMatrix(internal);

  return output;
}
