/********************************************************************* 
-- S+ sparse GE solver (version 1.1) --

Copyright (C) 1999 by University of California at Santa Barbara. 
Copyright (C) 2005 by University of Rochester. 
All rights reserved.

S+ team contact: Kai Shen <kshen@cs.rochester.edu>, 
                 Tao Yang <tyang@cs.ucsb.edu>  

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose, without fee, and without written agreement
is hereby granted, provided that the above copyright notice and the
following paragraph appear in all copies of this software.

This software is provided ``as is'', without any warranty, including 
any implied warranty for merchantability or fitness for a particular
purpose. Under no circumstances shall the University of California at 
Santa Barbara, the University of Rochester, or their agents be liable 
for any use of, misuse of, or inability to use this software, including 
incidental and consequential damages.
************************************************************************/

/*
 * Utils.c - miscellaneous utility routines
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <math.h>
#include <sys/time.h>

#include "struct.h"
#include "functions.h"
#include "globals.h"

/* Wall clock time */
double SP_Wtime() {
  struct timeval t;
  gettimeofday(&t, NULL);
  return (double)t.tv_sec+t.tv_usec*0.000001;
}

#ifdef _TRACKSPACE_

static int _TotalSpace = 0;

/* memory allocation */
void *SP_Malloc(size_t size) {
  void *mp;

  if (size == 0) return NULL;

  mp = malloc(size);
  if (mp == NULL) 
    SP_err(SP_ERR_LOC, "Error: malloc(%d) out of memory! (%d bytes allocated)", size, _TotalSpace);

  _TotalSpace += size;
  return mp;
}

/* memory free */
void SP_Free(void *addr) {
  free (addr);
}

/* memory free with size specified */
void SP_FreeSize(void *addr, int size) {
  free (addr);
  /*  _TotalSpace -= size; */
}

int SP_TotalSpace() {
  return _TotalSpace;
}

#else

/* memory allocation */
void *SP_Malloc(size_t size) {
  void *mp;
  if (size==0) return NULL;
  if ((mp=malloc(size)) == NULL) 
    SP_err(SP_ERR_LOC, "Error: malloc(%d) out of memory!", size);
  return mp;
}

#endif  /* _TRACKSPACE_ */

/* memory allocation for a chunk of elements */
ELEMTYPE *SP_MallocElem(size_t numElem) {
  size_t size = numElem * sizeof(ELEMTYPE);
  void *mp;
 
  mp = SP_Malloc(size);
  if (mp != NULL) bzero(mp, size);

  return mp;
}

#define SP_IO_ANY 1	/* enable IO for all MPI nodes */

/* default error handler */
void SP_err(char *file, int line, char *format, ...) {
  int flag;
  MPI_Initialized(&flag);
  if (flag) {
    int *ioproc;
    MPI_Attr_get(MPI_COMM_WORLD, MPI_IO, &ioproc, &flag);
    if (SP_IO_ANY || (flag && (*ioproc==MPI_ANY_SOURCE || *ioproc==SP_me_no))) {
      va_list argp;
      fprintf(stderr, "A fatal error occurred in file %s, line %d on processor %d:\n",
	      file, line, SP_me_no);
      va_start(argp, format);
      vfprintf(stderr, format, argp);
      va_end(argp);
      fprintf(stderr, "\n");
    }
    MPI_Abort(MPI_COMM_WORLD, -1);
  }
  else {
    va_list argp;
    fprintf(stderr, "A fatal error occurred in file %s, line %d on processor %d:\n",
	    file, line, SP_me_no);
    va_start(argp, format);
    vfprintf(stderr, format, argp);
    va_end(argp);
    fprintf(stderr, "\n");
    exit(-1);
  }
}

int SP_BlkSize = 28;	                        /* block size limit */
double SP_Relax = 0.3;				/* relax parameter for amalgamation */
double SP_DenseThresh = 0.85;			/* threshhold for a block to be considered dense */
double SP_Epsilon = 0;				/* a tiny number */
double SP_EpsilonBPratio = 1e-03;		/* a ratio for batch pivot */

/* half-precision error
#ifdef SINGLE_PRECISION
double SP_Epsilon = 3.45267e-04;
#else
double SP_Epsilon = 1.49012e-08;
#endif
*/

int 
SP_SetParam_BlkSize(
	int size	/* INPUT - block size limit */
	) 
/*
 * Purpose
 * =======
 *
 * Set block size limit which the maximum size a block can be.
 * This value affects cache efficiency and it should be set properly to allow 3 blocks
 * coexist in the first level cache at the same time, the default value is 25.
 * Returns 1 if successful, 0 otherwise.
 */
{
  if (size < 1) return 0;
  SP_BlkSize = size;
  return 1;
}


int 
SP_GetParam_BlkSize() 
/*
 * Purpose
 * =======
 *
 * Get block size limit.
 * Please refer to SP_SetParam_BlkSize for the description of the block size limit.
 */
{
  return SP_BlkSize;
}


int 
SP_SetParam_Relax(
	double relax	/* INPUT - relax parameter for amalgamation */
	) 
/*
 * Purpose
 * =======
 *
 * Set relax parameter for amalgamation which is the maximum extra fill-in allowed to generated
 * by amalgamation.
 * The value should be big enough to increase the average block size while it should be small
 * enough to reduce the number of extra fill-ins. The default value is 0.3 which means at most
 * 30% per cent extra fill-ins are allowed in amalgamation.
 * Returns 1 if successful, 0 otherwise.
 */
{
  if (relax < 0.0) return 0;
  SP_Relax = relax;
  return 1;
}


double 
SP_GetParam_Relax() 
/*
 * Purpose
 * =======
 *
 * Get relax parameter for amalgamation.
 * Please refer to SP_SetParam_Relax for the description of the relax parameter.
 */
{
  return SP_Relax;
}


int 
SP_SetParam_ThrholdPivot(
	double thrhold	/* INPUT - parameter for threshold pivoting */
	) 
/*
 * Purpose
 * =======
 *
 * Set the parameter for threshold pivoting.
 * Threshold pivoting allows the pivot choice to be other than the largest element in 
 * the pivot column, as long as it is within a certain fraction (u <= 1.0) of the 
 * largest element.  A smaller u allows more freedom in pivot selection, however, it
 * might also weaken the numerical stability of LU factorization.
 * Returns 1 if successful, 0 otherwise.
 */
{
  if (thrhold < 0.0 || thrhold > 1.0) return 0;
  SP_PivotThresh = thrhold;
  return 1;
}


double 
SP_GetParam_ThrholdPivot() 
/*
 * Purpose
 * =======
 *
 * Get the parameter for threshold pivoting.
 * Please refer to SP_SetParam_ThrholdPivot for the description of the parameter.
 */
{
  return SP_PivotThresh;
}


int 
SP_SetParam_BatchPivot(
	int batchpivot	/* INPUT - parameter for batch pivoting */
	) 
/*
 * Purpose
 * =======
 *
 * Set the parameter for batch pivoting.
 * Batch pivoting can reduce the synchronization frequency during pivot selection.
 * There are three options for this parameter: SP_PIVOTING_COLBYCOL means 
 * column-by-column pivoting (or no batch pivoting), SP_PIVOTING_LD means
 * large diagonal batch pivoting; SP_PIVOTING_SBP means speculative batch
 * pivoting.
 * Returns 1 if successful, 0 otherwise.
 */
{
  if (batchpivot < 0 || batchpivot >= SP_PIVOTING_COUNT) return 0;
  SP_BatchPivot = batchpivot;
  return 1;
}


int
SP_GetParam_BatchPivot() 
/*
 * Purpose
 * =======
 *
 * Get the parameter for batch pivoting.
 * Please refer to SP_SetParam_BatchPivot for the description of the parameter.
 */
{
  return SP_BatchPivot;
}


#if 0
/* set dense block threshhold, returns 1 if successful, 0 otherwise */
int SP_SetParam_Dense(double thresh) {
  if (thresh < 0 || thresh > 1) return 0;
  SP_DenseThresh = thresh;
  return 1;
}

/* get dense block threshhold */
double SP_GetParam_Dense() {
  return SP_DenseThresh;
}
#endif

/* calculate the order of bits in val */
int SP_OrderOfInt(PATTYPE val) {
  register int i=0;

  while ( val ) {
    i += val & 1;
    val >>= 1;
  }
  return i;
}

SP_SparseMatrix*
SP_ReadMatrixFromIJ (
	const char *fname	/* INPUT - name of input file */
	) 
/*
 * Purpose
 * =======
 *
 * generate a input matrix from a ij-value data file
 */
{
  FILE *file;
  int l1, l2, l3, cur, lastCol;
  double d;
  SP_SparseMatrix *ma;

  if ((file = fopen(fname, "r")) == NULL)
    SP_err(SP_ERR_LOC, "File open error");
  fscanf(file, "%d %d %d", &l1, &l2, &l3);
  while (getc(file) != '\n');

  ma = (SP_SparseMatrix*) SP_Malloc(sizeof(SP_SparseMatrix));
  ma->order = l1;
  ma->numNZ = l3;
  ma->coe = (ELEMTYPE*) SP_Malloc(sizeof(ELEMTYPE)*(ma->numNZ));
  ma->rowIndex = (int*) SP_Malloc(sizeof(int)*(ma->numNZ));
  ma->colStart = (int*) SP_Malloc(sizeof(int)*(ma->order+1));

  cur = 0;
  lastCol = 0;
  for (;;) {
    char buffer[256];
    char *pl2, *pd;

    fgets(buffer, 256, file);
    pl2=strchr(buffer, ' ')+1;
    pd=strchr(pl2, ' ')+1;
    l1=atoi(buffer);
    l2=atoi(pl2);
    d=strtod(pd, NULL);

    if (l1 == 0) break;

    ma->coe[cur] = d;
    /*
    ma->rowIndex[cur] = l1;
    */
    ma->rowIndex[cur] = l1-1;

    if (l2 > lastCol) {
      lastCol++;
      if (l2 != lastCol)
	SP_err(SP_ERR_LOC, "Illegal input matrix");
      ma->colStart[l2-1]=cur;
    }
    
    cur++;
  }
  ma->colStart[ma->order] = ma->numNZ;

  fclose(file);
  return ma;
}


/* helpers for SP_ReadMatrixFromHB() */

#define HB_MAXLINE 256

inline int _ToInt(char *buf, int buflen, char *tmpbuf) {
  memcpy(tmpbuf, buf, buflen);
  tmpbuf[buflen] = '\0';
  return atoi(tmpbuf);
}

inline double _ToReal(char *buf, int buflen, char *tmpbuf) {
  int with_eord=0, with_dp=0;
  int i;
  memcpy(tmpbuf, buf, buflen);
  tmpbuf[buflen] = '\0';
  for (i=0; i<buflen; i++) {
    if (tmpbuf[i]=='e' || tmpbuf[i]=='E') with_eord=1;
    if (tmpbuf[i]=='d' || tmpbuf[i]=='D') {
      tmpbuf[i]='E';
      with_eord=1;
    }
    if (tmpbuf[i]=='.') with_dp=1;
  }
  /* must include 'd/D/e/E' and decimal point to avoid worrying about any scaling issue */
  if (with_eord!=1 || with_dp!=1)
    SP_err(SP_ERR_LOC, "Illegal input matrix");

  return atof(tmpbuf);
}

/* only support (d+[\I|\i]d+) format */
void _InputFmtInt(char *buf, int buflen, int *fmtcnt, int *fmtlen, char *tmpbuf)
{
  int start, i_mark, end;

  for (start=0; start<buflen && buf[start]!='('; start++);
  if (start>=buflen)
    SP_err(SP_ERR_LOC, "Illegal input matrix");

  for (i_mark=start+1; i_mark<buflen && buf[i_mark]!='i' && buf[i_mark]!='I'; i_mark++);
  if (i_mark>=buflen)
    SP_err(SP_ERR_LOC, "Illegal input matrix");

  for (end=i_mark+1; end<buflen && buf[end]!=')'; end++);
  if (end>=buflen)
    SP_err(SP_ERR_LOC, "Illegal input matrix");

  *fmtcnt = _ToInt(buf+start+1, i_mark-start-1, tmpbuf);
  *fmtlen = _ToInt(buf+i_mark+1, end-i_mark-1, tmpbuf);
}

/* only support (d+[\D|\d|\E|\e|\F|\f|\G|\g]d+\.d+), one exception is that there is possibly \p|\P in the beginning */
void _InputFmtReal(char *buf, int buflen, int *fmtcnt, int *fmtlen, char *tmpbuf)
{
  int start, r_mark, dp_mark, end;

  for (start=0; start<buflen && buf[start]!='('; start++);
  if (start>=buflen)
    SP_err(SP_ERR_LOC, "Illegal input matrix");

  for (r_mark=start+1; r_mark<buflen && buf[r_mark]!='D' && buf[r_mark]!='d' && buf[r_mark]!='E' && buf[r_mark]!='e' && buf[r_mark]!='F' && buf[r_mark]!='f' && buf[r_mark]!='G' && buf[r_mark]!='g'; r_mark++) {
    if (buf[r_mark]=='p' || buf[r_mark]=='P') {
      start = r_mark;
    }
  }
  if (r_mark>=buflen)
    SP_err(SP_ERR_LOC, "Illegal input matrix");

  for (dp_mark=r_mark+1; dp_mark<buflen && buf[dp_mark]!='.'; dp_mark++);
  if (dp_mark>=buflen)
    SP_err(SP_ERR_LOC, "Illegal input matrix");

  for (end=dp_mark+1; end<buflen && buf[end]!=')'; end++);
  if (end>=buflen)
    SP_err(SP_ERR_LOC, "Illegal input matrix");

  *fmtcnt = _ToInt(buf+start+1, r_mark-start-1, tmpbuf);
  *fmtlen = _ToInt(buf+r_mark+1, dp_mark-r_mark-1, tmpbuf);
}


SP_SparseMatrix*
SP_ReadMatrixFromHB (
	const char *fname	/* INPUT - name of input file */
	) 
/*
 * Purpose
 * =======
 *
 * generate a input matrix from a HB format sparse matrix data file
 */
{
  FILE *file = NULL;
  SP_SparseMatrix *ma = NULL;
  int line_tot=0, line_colstart=0, line_rowidx=0, line_coe=0, line_rhs=0;
  int numrow=0, numcol=0, numnz=0, numeltma=0;
  int fmt_colstart_cnt=0, fmt_colstart_len=0, fmt_rowidx_cnt=0, fmt_rowidx_len=0, fmt_coe_cnt=0, fmt_coe_len=0;
  int i, j, k;

  file = fopen(fname, "r");
  if (file==NULL)
    SP_err(SP_ERR_LOC, "File open error");

  char linebuf[HB_MAXLINE+1];
  char tmpbuf[HB_MAXLINE+1];

  /* parse header line 1 */
  fgets(linebuf, HB_MAXLINE, file);
  /* ignore title and key */

  /* parse header line 2 */
  fgets(linebuf, HB_MAXLINE, file);
  line_tot = _ToInt(linebuf, 14, tmpbuf);
  line_colstart = _ToInt(linebuf+14, 14, tmpbuf);
  line_rowidx = _ToInt(linebuf+28, 14, tmpbuf);
  line_coe = _ToInt(linebuf+42, 14, tmpbuf);
  line_rhs = _ToInt(linebuf+56, 14, tmpbuf);
  if (line_tot != line_colstart+line_rowidx+line_coe+line_rhs)
    SP_err(SP_ERR_LOC, "Illegal input matrix");

  /* parse header line 3 */
  fgets(linebuf, HB_MAXLINE, file);
  /* only support RUA */
  if (linebuf[0]!='R' || linebuf[1]!='U' || linebuf[2]!='A')
    SP_err(SP_ERR_LOC, "Illegal input matrix");
  numrow = _ToInt(linebuf+14, 14, tmpbuf);
  numcol = _ToInt(linebuf+28, 14, tmpbuf);
  numnz = _ToInt(linebuf+42, 14, tmpbuf);
  numeltma = _ToInt(linebuf+56, 14, tmpbuf);
  if (numrow!=numcol)
    SP_err(SP_ERR_LOC, "Illegal input matrix");
  /* only support assembled matrix */
  if (numeltma!=0)
    SP_err(SP_ERR_LOC, "Illegal input matrix");    

  /* parse header line 4 */
  fgets(linebuf, HB_MAXLINE, file);
  _InputFmtInt(linebuf, 16, &fmt_colstart_cnt, &fmt_colstart_len, tmpbuf);
  _InputFmtInt(linebuf+16, 16, &fmt_rowidx_cnt, &fmt_rowidx_len, tmpbuf);
  _InputFmtReal(linebuf+32, 20, &fmt_coe_cnt, &fmt_coe_len, tmpbuf);
  if ((fmt_colstart_cnt*fmt_colstart_len > 80) || (fmt_rowidx_cnt*fmt_rowidx_len > 80) || (fmt_coe_cnt*fmt_coe_len > 80))
    SP_err(SP_ERR_LOC, "Illegal input matrix");

  /* parse header line 5 if exists */
  if (line_rhs>0) {
    fgets(linebuf, HB_MAXLINE, file);
    /* ignore the right-hand side */
  }
  
  /* initialize the matrix */
  ma = (SP_SparseMatrix*) malloc(sizeof(SP_SparseMatrix));
  if (ma==NULL) 
    SP_err(SP_ERR_LOC, "Out of memory");
  ma->order = numrow;
  ma->numNZ = numnz;
  ma->coe = (ELEMTYPE*) malloc(sizeof(ELEMTYPE)*(ma->numNZ));
  ma->rowIndex = (int*) malloc(sizeof(int)*(ma->numNZ));
  ma->colStart = (int*) malloc(sizeof(int)*(ma->order+1));
  if (ma->coe==NULL || ma->rowIndex==NULL || ma->colStart==NULL)
    SP_err(SP_ERR_LOC, "Out of memory");
  
  /* read the column start */
  k = 0;
  for (i=0; i<line_colstart; i++) {
    fgets(linebuf, HB_MAXLINE, file);

    for (j=0; j<fmt_colstart_cnt; j++) {
      ma->colStart[k] = _ToInt(linebuf+(k%fmt_colstart_cnt)*fmt_colstart_len, fmt_colstart_len, tmpbuf) - 1;
      if (ma->colStart[k]>ma->numNZ)
	SP_err(SP_ERR_LOC, "Illegal input matrix");
      if (k>0 && ma->colStart[k]<ma->colStart[k-1])
	SP_err(SP_ERR_LOC, "Illegal input matrix");
      k++;
      if (k>=ma->order+1) break;
    }
  }
  if (ma->colStart[ma->order]!=ma->numNZ)
    SP_err(SP_ERR_LOC, "Illegal input matrix");
  
  /* read the row index */
  k = 0;
  for (i=0; i<line_rowidx; i++) {
    fgets(linebuf, HB_MAXLINE, file);

    for (j=0; j<fmt_rowidx_cnt; j++) {
      ma->rowIndex[k] = _ToInt(linebuf+(k%fmt_rowidx_cnt)*fmt_rowidx_len, fmt_rowidx_len, tmpbuf) - 1;
      if (ma->rowIndex[k]>=ma->order)
	SP_err(SP_ERR_LOC, "Illegal input matrix");
      k++;
      if (k>=ma->numNZ) break;
    }
  }

  /* read the coe */
  k = 0;
  for (i=0; i<line_coe; i++) {
    fgets(linebuf, HB_MAXLINE, file);

    for (j=0; j<fmt_coe_cnt; j++) {
      ma->coe[k] = _ToReal(linebuf+(k%fmt_coe_cnt)*fmt_coe_len, fmt_coe_len, tmpbuf);
      k++;
      if (k>=ma->numNZ) break;
    }
  }

  fclose(file);
  return ma;
}


/* #define _1DC_MAPPING_ */
/* #define _1DR_MAPPING_ */

int SP_me_no, SP_nproc;		/* my PE id and total PE numbers */
int SP_rp, SP_cp;		/* the processors are viewed as a 2-D mesh SP_rp x SP_cp*/
int SP_me_rno, SP_me_cno;	/* the 2-D coordinates of this processor   */

/* determine a 2-D mapping of a linear array of processors */
static void _2Dmapping(int me_no, int nproc) {
  SP_me_no = me_no;
  SP_nproc = nproc;

#if defined(_1DC_MAPPING_)
  SP_rp = 1;
  SP_cp = nproc/SP_rp;
#elif defined(_1DR_MAPPING_)
  SP_cp = 1;
  SP_rp = nproc/SP_cp;
#else
  SP_rp = (int) sqrt(nproc);
  while(nproc % SP_rp) --SP_rp;
  /* if (!(SP_rp%2) &&  nproc/SP_rp==SP_rp) SP_rp/=2; */
  SP_cp=nproc/SP_rp;
#endif

  SP_me_rno=me_no%SP_rp;
  SP_me_cno=me_no/SP_rp;
}

/* Symbolic factorization, allocate space and initialize data */
void SP_Initialize(SP_SparseMatrix *A, int me_no, int nproc, int *perm_c, int *perm_r, SP_SNmatrix_struct **SNma) {
  Matrix *ma1 = NULL, *ma2 = NULL;
  Node **eforest = NULL;       			/* element-wise elimination forest */

  _2Dmapping(me_no, nproc);			/* initialize the 2D processor grid */

  *SNma = (SP_SNmatrix_struct*) SP_Malloc (sizeof(SP_SNmatrix_struct));

  if (me_no==0) {
    eforest = (Node**) SP_Malloc((A->order+1) * sizeof(Node*));
    ma1 = SP_SymFac(A, eforest, perm_c, perm_r);/* symbolic factorization */
    (*SNma)->eleOrder = ma1->order;
    SP_SnodePA_p(*SNma, ma1, eforest, SP_Relax);	/* supernode partitioning and amalgamation */
    ma2 = SP_SnodeMatrix(*SNma, ma1);		/* generate supernode matrix */
  }

  SP_AllocIndex1(*SNma);			/* allocate space for index part 1 */

  if (me_no==0) 
    SP_GenIndex1(*SNma, ma2, ma1);		/* generate the index part 1 */

  SP_BcastIndex1(*SNma);		      	/* broadcast data in index part 1 */
  SP_GenIndex2(*SNma);				/* generate the index part 2 */
  SP_GenSnodeChild(*SNma, eforest);		/* generate the SnodeChild array */

  if (me_no==0) {
    SP_FreeEforest(eforest, A->order);
    SP_FreeMatrix(ma1);
    SP_FreeMatrix(ma2);
  }

  SP_InitData(*SNma, A, perm_c, perm_r);      	/* allocate and initialize buffers and data in the matrix */
}

void 
SP_Solve (
	SP_SparseMatrix *A,	/* INPUT - sparse matrix */
	SP_DenseMatrix *B,	/* INPUT/OUTPUT - right-hand size dense matrix, 
				   will be overwritten by result matrix on exit */
	int *perm_c,		/* INPUT - column permutation vector */
	int *perm_r,		/* INPUT - row permutation vector */
	int me_no,		/* INPUT - local MPI node ID */
	int nproc		/* INPUT - total MPI nodes number */
) 
/*
 * Purpose
 * =======
 *
 * Solve linear system A*X=B, the storage of B will be overwritten by result X.
 */
{
  SP_SNmatrix_struct *SNma;
  rowID *Prow;			/* row permutation array for pivoting */

  SP_Initialize(A, me_no, nproc, perm_c, perm_r, &SNma);
  SP_LUP(SNma, &Prow);
  SP_TR_Solve(SNma, Prow, B, perm_c, perm_r);

  SP_Free(Prow);
  SP_DestroySNmatrix(SNma);
}


SP_DenseMatrix *SP_GenXtrue(int colNum, int rowNum) {
  SP_DenseMatrix *Xtrue;
  int i;

  Xtrue = (SP_DenseMatrix*) SP_Malloc (sizeof(SP_DenseMatrix));
  Xtrue->colNum = colNum;
  Xtrue->rowNum = rowNum;
  Xtrue->element = (ELEMTYPE*) SP_Malloc (colNum*rowNum*sizeof(SP_DenseMatrix));
  for (i=0; i<colNum*rowNum; i++) Xtrue->element[i] = i + 1;

  return Xtrue;
}

SP_DenseMatrix *SP_GenRHS(SP_SparseMatrix *A, SP_DenseMatrix *Xtrue) {
  SP_DenseMatrix *B;
  register int i, j, k, h;

  B = (SP_DenseMatrix*) SP_Malloc (sizeof(SP_DenseMatrix));
  B->colNum = Xtrue->colNum;
  B->rowNum = Xtrue->rowNum;
  B->element = (ELEMTYPE*) SP_Malloc (B->colNum * B->rowNum * sizeof(ELEMTYPE));
  bzero( B->element, B->colNum * B->rowNum * sizeof(ELEMTYPE));

  for (i=0; i<A->order; i++)
    for (j=A->colStart[i]; j<A->colStart[i+1]; j++)
      for (k=0; k<Xtrue->colNum; k++) {
	h = A->rowIndex[j];
	B->element[k*A->order+h] += A->coe[j] * Xtrue->element[k*A->order+i];
      }

  return B;
}

/* get matrix norm */
void SP_GetMatrixNorm(SP_SparseMatrix *ma, ELEMTYPE *max, ELEMTYPE *tot) {
  int i;
  ELEMTYPE x;

  if (ma==NULL) return;

  *max = 0;
  *tot = 0;
  for (i=0; i<ma->numNZ; i++) {
    x = fabs(ma->coe[i]);
    if (x>*max) *max = x;
    *tot += x;
  }
}

void SP_NormError(SP_DenseMatrix *X, SP_DenseMatrix *Xtrue) {
  int i, j;
  ELEMTYPE err, xnorm;

  for (i=0; i<X->colNum; i++) {
    err = xnorm = 0.0;
    for (j=0; j<X->rowNum; j++) {
      err = MAX(err, fabs(X->element[i*X->rowNum+j] - Xtrue->element[i*X->rowNum+j]));
      xnorm = MAX(xnorm, fabs(Xtrue->element[i*X->rowNum+j]));
    }
    err = err / xnorm;
    fprintf(stdout, "||X - Xtrue||/||Xtrue|| = %e\n", err);
  }
}

void SP_PrintDenseMatrix(SP_DenseMatrix *M) {
  int i, j;

  for (i=0; i<M->colNum; i++) {
    fprintf(stdout, "Column %d:", i);
    for (j=0; j<M->rowNum; j++) 
      fprintf(stdout, " %3.3f", M->element[i*M->rowNum+j]);
    fprintf(stdout, "\n");
  }
}

/* Release the memory space occupied by a sparse matrix */
void SP_DestroySparseMatrix(SP_SparseMatrix *ma) {
  SP_FreeSize(ma->coe, sizeof(ELEMTYPE)*(ma->numNZ));
  SP_FreeSize(ma->rowIndex, sizeof(int)*(ma->numNZ));
  SP_FreeSize(ma->colStart, sizeof(int)*(ma->order+1));
  SP_FreeSize(ma, sizeof(SP_SparseMatrix));
}

/* return the value of entry (row, col) */
static ELEMTYPE _SparseMatrixEntryValue(SP_SparseMatrix *SM, int row, int col) {
  ELEMTYPE eleval = 0.0;
  int idx, row2;
  for (idx=SM->colStart[col]; idx<SM->colStart[col+1]; idx++) {
    row2 = SM->rowIndex[idx];
    if (row2==row) {
      eleval = SM->coe[idx];
      break;
    } 
    /* comment the following if the indexes in rowIndex arrays are not sorted */
    if (row2>row) break;
  }
  return eleval;
}

int SP_SparseMatrixDiagonalZero(SP_SparseMatrix *SM) {
  int k;
  int count = 0;

  for (k=0; k<SM->order; k++)
    if (_SparseMatrixEntryValue(SM, k, k)==0.0)
      count ++;

  return count;
}
